@@ -1,6 +1,7 @@
 from itertools import chain
 
 import numpy as np
+import scipy.sparse as sp
 
 from AnyQt.QtCore import Qt, QAbstractTableModel
 from AnyQt.QtGui import QColor
@@ -8,6 +9,7 @@ from AnyQt.QtWidgets import QComboBox, QTableView, QSizePolicy
 
 from Orange.data import DiscreteVariable, ContinuousVariable, StringVariable, \
     TimeVariable, Domain
+from Orange.statistics.util import unique
 from Orange.widgets import gui
 from Orange.widgets.gui import HorizontalGridDelegate
 from Orange.widgets.settings import ContextSetting
@@ -33,7 +35,7 @@ class VarTableModel(QAbstractTableModel):
     DISCRETE_VALUE_DISPLAY_LIMIT = 20
 
     places = "feature", "target", "meta", "skip"
-    typenames = "nominal", "numeric", "string", "datetime"
+    typenames = "categorical", "numeric", "string", "datetime"
     vartypes = DiscreteVariable, ContinuousVariable, StringVariable, TimeVariable
     name2type = dict(zip(typenames, vartypes))
     type2name = dict(zip(vartypes, typenames))
@@ -79,7 +81,7 @@ class VarTableModel(QAbstractTableModel):
         row, col = index.row(), index.column()
         row_data = self.variables[row]
         if role == Qt.EditRole:
-            if col == Column.name:
+            if col == Column.name and not (value.isspace() or value == ""):
                 row_data[col] = value
             elif col == Column.tpe:
                 vartype = self.name2type[value]
@@ -125,7 +127,7 @@ class ComboDelegate(HorizontalGridDelegate):
             def hidePopup(me):
                 if me.popup_shown:
                     self.view.model().setData(
-                            index, me.highlighted_text, Qt.EditRole)
+                        index, me.highlighted_text, Qt.EditRole)
                     self.popup_shown = False
                 super().hidePopup()
                 self.view.closeEditor(me, self.NoHint)
@@ -140,9 +142,15 @@ class VarTypeDelegate(ComboDelegate):
         combo.clear()
         no_numeric = not self.view.model().variables[
             index.row()][Column.not_valid]
-        ind = self.items.index(index.data())
-        combo.addItems(self.items[:1] + self.items[1 + no_numeric:])
-        combo.setCurrentIndex(ind - (no_numeric and ind > 1))
+        if no_numeric:
+            # Do not allow selection of numeric and datetime
+            items = [i for i in self.items if i not in ("numeric", "datetime")]
+        else:
+            items = self.items
+
+        ind = items.index(index.data())
+        combo.addItems(items)
+        combo.setCurrentIndex(ind)
 
 
 class PlaceDelegate(ComboDelegate):
@@ -190,6 +198,37 @@ class DomainEditor(QTableView):
         self.place_delegate = PlaceDelegate(self, VarTableModel.places)
         self.setItemDelegateForColumn(Column.place, self.place_delegate)
 
+    @staticmethod
+    def _is_missing(x):
+        return str(x) in ("nan", "")
+
+    @staticmethod
+    def _iter_vals(x):
+        """Iterate over values of sparse or dense arrays."""
+        for i in range(x.shape[0]):
+            yield x[i, 0]
+
+    @staticmethod
+    def _to_column(x, to_sparse, dtype=None):
+        """Transform list of values to sparse/dense column array."""
+        x = np.array(x, dtype=dtype).reshape(-1, 1)
+        if to_sparse:
+            x = sp.csc_matrix(x)
+        return x
+
+    @staticmethod
+    def _merge(cols, force_dense=False):
+        if len(cols) == 0:
+            return None
+
+        all_dense = not any(sp.issparse(c) for c in cols)
+        if all_dense:
+            return np.hstack(cols)
+        if force_dense:
+            return np.hstack([c.toarray() if sp.issparse(c) else c for c in cols])
+        sparse_cols = [c if sp.issparse(c) else sp.csc_matrix(c) for c in cols]
+        return sp.hstack(sparse_cols).tocsr()
+
     def get_domain(self, domain, data):
         """Create domain (and dataset) from changes made in the widget.
 
@@ -202,27 +241,29 @@ class DomainEditor(QTableView):
         -------
         (new_domain, [attribute_columns, class_var_columns, meta_columns])
         """
+        # Allow type-checking with type() instead of isinstance() for exact comparison
+        # pylint: disable=unidiomatic-typecheck
+
         variables = self.model().variables
         places = [[], [], []]  # attributes, class_vars, metas
         cols = [[], [], []]  # Xcols, Ycols, Mcols
 
-        def is_missing(x):
-            return str(x) in ("nan", "")
-
         for (name, tpe, place, _, _), (orig_var, orig_plc) in \
                 zip(variables,
-                    chain([(at, Place.feature) for at in domain.attributes],
-                          [(cl, Place.class_var) for cl in domain.class_vars],
-                          [(mt, Place.meta) for mt in domain.metas])):
+                        chain([(at, Place.feature) for at in domain.attributes],
+                              [(cl, Place.class_var) for cl in domain.class_vars],
+                              [(mt, Place.meta) for mt in domain.metas])):
             if place == Place.skip:
                 continue
-            if orig_plc == Place.meta:
-                col_data = data[:, orig_var].metas
-            elif orig_plc == Place.class_var:
-                col_data = data[:, orig_var].Y
-            else:
-                col_data = data[:, orig_var].X
-            col_data = col_data.ravel()
+
+            col_data = self._get_column(data, orig_var, orig_plc)
+            is_sparse = sp.issparse(col_data)
+
+            cont_ints = type(orig_var) == ContinuousVariable and \
+                        all(x.is_integer() for x in self._iter_vals(col_data) if not np.isnan(x))
+            disc_ints = type(orig_var) == DiscreteVariable and \
+                        all(x.isdecimal() for x in orig_var.values)
+
             if name == orig_var.name and tpe == type(orig_var):
                 var = orig_var
             elif tpe == type(orig_var):
@@ -230,20 +271,53 @@ class DomainEditor(QTableView):
                 orig_var.name = name
                 var = orig_var
             elif tpe == DiscreteVariable:
-                values = list(str(i) for i in np.unique(col_data) if not is_missing(i))
+                values = list(str(i) for i in unique(col_data) if not self._is_missing(i))
+                col_data = [np.nan if self._is_missing(x) else values.index(str(x))
+                            for x in self._iter_vals(col_data)]
+                if cont_ints:
+                    values = [str(int(float(v))) for v in values]
                 var = tpe(name, values)
-                col_data = [np.nan if is_missing(x) else values.index(str(x))
-                            for x in col_data]
-            elif tpe == StringVariable and type(orig_var) == DiscreteVariable:
+                col_data = self._to_column(col_data, is_sparse)
+            elif tpe == StringVariable:
                 var = tpe(name)
-                col_data = [orig_var.repr_val(x) if not np.isnan(x) else ""
-                            for x in col_data]
+                if type(orig_var) == DiscreteVariable:
+                    col_data = [orig_var.repr_val(x) if not np.isnan(x) else ""
+                                for x in self._iter_vals(col_data)]
+                elif type(orig_var) == ContinuousVariable:
+                    col_data = [str(int(x)) if cont_ints else orig_var.repr_val(x)
+                                if not np.isnan(x) else ""
+                                for x in self._iter_vals(col_data)]
+                # don't obey sparsity for StringVariable since they are
+                # in metas which are transformed to dense below
+                col_data = self._to_column(col_data, False, dtype=object)
+            elif tpe == ContinuousVariable and type(orig_var) == DiscreteVariable:
+                var = tpe(name)
+                if disc_ints:
+                    col_data = [np.nan if self._is_missing(x) else float(orig_var.values[int(x)])
+                                for x in self._iter_vals(col_data)]
+                col_data = self._to_column(col_data, is_sparse)
             else:
                 var = tpe(name)
             places[place].append(var)
             cols[place].append(col_data)
+
+        # merge columns for X, Y and metas
+        feats = cols[Place.feature]
+        X = self._merge(feats) if len(feats) else np.empty((len(data), 0))
+        Y = self._merge(cols[Place.class_var], force_dense=True)
+        m = self._merge(cols[Place.meta], force_dense=True)
         domain = Domain(*places)
-        return domain, cols
+        return domain, [X, Y, m]
+
+    def _get_column(self, data, source_var, source_place):
+        """ Extract column from data and preserve sparsity. """
+        if source_place == Place.meta:
+            col_data = data[:, source_var].metas
+        elif source_place == Place.class_var:
+            col_data = data[:, source_var].Y.reshape(-1, 1)
+        else:
+            col_data = data[:, source_var].X
+        return col_data
 
     def set_domain(self, domain):
         self.variables = self.parse_domain(domain)
@@ -277,7 +351,8 @@ class DomainEditor(QTableView):
             return False
 
         def discrete_value_display(value_list):
-            result = ", ".join(str(v) for v in value_list[:VarTableModel.DISCRETE_VALUE_DISPLAY_LIMIT])
+            result = ", ".join(str(v)
+                               for v in value_list[:VarTableModel.DISCRETE_VALUE_DISPLAY_LIMIT])
             if len(value_list) > VarTableModel.DISCRETE_VALUE_DISPLAY_LIMIT:
                 result += ", ..."
             return result
