@@ -1,18 +1,26 @@
+# Module imports Input, Output and AttributeList to be used in widgets
+# pylint: disable=unused-import
+
 import sys
 import os
 import types
 
 from AnyQt.QtWidgets import (
     QWidget, QDialog, QVBoxLayout, QSizePolicy, QApplication, QStyle,
-    QShortcut, QSplitter, QSplitterHandle, QPushButton
+    QShortcut, QSplitter, QSplitterHandle, QPushButton, QStatusBar,
+    QProgressBar
 )
 from AnyQt.QtCore import (
-    Qt, QByteArray, QSettings, QUrl, pyqtSignal as Signal)
+    Qt, QByteArray, QSettings, QUrl, pyqtSignal as Signal
+)
 from AnyQt.QtGui import QIcon, QKeySequence, QDesktopServices
 
 from Orange.data import FileFormat
 from Orange.widgets import settings, gui
-from Orange.canvas.registry import description as widget_description
+# OutputSignal and InputSignal are imported for compatibility, but shouldn't
+# be used; use Input and Output instead
+from Orange.canvas.registry import description as widget_description, \
+    WidgetDescription, OutputSignal, InputSignal
 from Orange.canvas.report import Report
 from Orange.widgets.gui import OWComponent
 from Orange.widgets.io import ClipboardFormat
@@ -20,8 +28,10 @@ from Orange.widgets.settings import SettingsHandler
 from Orange.widgets.utils import saveplot, getdeepattr
 from Orange.widgets.utils.progressbar import ProgressBarMixin
 from Orange.widgets.utils.messages import \
-    WidgetMessagesMixin, UnboundMsg
-from .utils.overlay import MessageOverlayWidget
+    WidgetMessagesMixin, UnboundMsg, MessagesWidget
+from Orange.widgets.utils.signals import \
+    WidgetSignalsMixin, Input, Output, AttributeList
+from Orange.widgets.utils.overlay import MessageOverlayWidget, OverlayWidget
 
 # Msg is imported and renamed, so widgets can import it from this module rather
 # than the one with the mixin (Orange.widgets.utils.messages). Assignment is
@@ -46,34 +56,18 @@ class WidgetMetaClass(type(QDialog)):
     #noinspection PyMethodParameters
     # pylint: disable=bad-classmethod-argument
     def __new__(mcs, name, bases, kwargs):
-        from Orange.canvas.registry.description import (
-            input_channel_from_args, output_channel_from_args)
-
         cls = super().__new__(mcs, name, bases, kwargs)
         if not cls.name: # not a widget
             return cls
-
-        cls.inputs = [input_channel_from_args(inp) for inp in cls.inputs]
-        cls.outputs = [output_channel_from_args(outp) for outp in cls.outputs]
-
-        for inp in cls.inputs:
-            if not hasattr(cls, inp.handler):
-                raise AttributeError("missing input signal handler '{}' in {}".
-                                     format(inp.handler, cls.name))
-
-        # TODO Remove this when all widgets are migrated to Orange 3.0
-        if (hasattr(cls, "settingsToWidgetCallback") or
-                hasattr(cls, "settingsFromWidgetCallback")):
-            raise TypeError("Reimplement settingsToWidgetCallback and "
-                            "settingsFromWidgetCallback")
-
-        cls.settingsHandler = SettingsHandler.create(cls, template=cls.settingsHandler)
-
+        cls.convert_signals()
+        cls.settingsHandler = \
+            SettingsHandler.create(cls, template=cls.settingsHandler)
         return cls
 
 
 class OWWidget(QDialog, OWComponent, Report, ProgressBarMixin,
-               WidgetMessagesMixin, metaclass=WidgetMetaClass):
+               WidgetMessagesMixin, WidgetSignalsMixin,
+               metaclass=WidgetMetaClass):
     """Base widget class"""
 
     # Global widget count
@@ -89,7 +83,7 @@ class OWWidget(QDialog, OWComponent, Report, ProgressBarMixin,
     version = None
     #: Short widget description (:class:`str` optional), displayed in
     #: canvas help tooltips.
-    description = None
+    description = ""
     #: Widget icon path relative to the defining module
     icon = "icons/Unknown.png"
     #: Widget priority used for sorting within a category
@@ -129,7 +123,7 @@ class OWWidget(QDialog, OWComponent, Report, ProgressBarMixin,
     #: responsible for displaying messages within the widget in an
     #: appropriate manner.
     want_message_bar = True
-    #: Widget painted by `Save graph" button
+    #: Widget painted by `Save graph` button
     graph_name = None
     graph_writers = FileFormat.img_writers
 
@@ -180,6 +174,7 @@ class OWWidget(QDialog, OWComponent, Report, ProgressBarMixin,
         QDialog.__init__(self, None, self.get_flags())
         OWComponent.__init__(self)
         WidgetMessagesMixin.__init__(self)
+        WidgetSignalsMixin.__init__(self)
 
         stored_settings = kwargs.get('stored_settings', None)
         if self.settingsHandler:
@@ -232,6 +227,21 @@ class OWWidget(QDialog, OWComponent, Report, ProgressBarMixin,
     def __init__(self, *args, **kwargs):
         """__init__s are called in __new__; don't call them from here"""
 
+    @classmethod
+    def get_widget_description(cls):
+        if not cls.name:
+            return
+        properties = {name: getattr(cls, name) for name in
+                      ("name", "icon", "description", "priority", "keywords",
+                       "help", "help_ref", "url",
+                       "version", "background", "replaces")}
+        properties["id"] = cls.id or cls.__module__
+        properties["inputs"] = cls.get_signals("inputs")
+        properties["outputs"] = cls.get_signals("outputs")
+        properties["qualified_name"] = \
+            "{}.{}".format(cls.__module__, cls.__name__)
+        return properties
+
     @classmethod
     def get_flags(cls):
         return (Qt.Window if cls.resizing_enabled
@@ -248,7 +258,7 @@ class OWWidget(QDialog, OWComponent, Report, ProgressBarMixin,
                 """Resize on left button"""
                 if event.button() == Qt.LeftButton:
                     splitter = self.splitter()
-                    splitter.setSizes([int(splitter.sizes()[0] == 0), 1000])
+                    splitter.setSizes([int(splitter.sizes()[0] == 0), 100000])
                 super().mouseReleaseEvent(event)
 
             def mouseMoveEvent(self, event):
@@ -315,21 +325,61 @@ class OWWidget(QDialog, OWComponent, Report, ProgressBarMixin,
         """
         self.setLayout(QVBoxLayout())
         self.layout().setContentsMargins(2, 2, 2, 2)
+
         if not self.resizing_enabled:
             self.layout().setSizeConstraint(QVBoxLayout.SetFixedSize)
 
         self.want_main_area = self.want_main_area or self.graph_name
         self._create_default_buttons()
 
-        if self.want_message_bar:
-            self.insert_message_bar()
-
         self._insert_splitter()
         if self.want_control_area:
             self._insert_control_area()
         if self.want_main_area:
             self._insert_main_area()
 
+        if self.want_message_bar:
+            # Use a OverlayWidget for status bar positioning.
+            c = OverlayWidget(self, alignment=Qt.AlignBottom)
+            c.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
+            c.setWidget(self)
+            c.setLayout(QVBoxLayout())
+            c.layout().setContentsMargins(0, 0, 0, 0)
+            sb = QStatusBar()
+            sb.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Maximum)
+            sb.setSizeGripEnabled(self.resizing_enabled)
+            c.layout().addWidget(sb)
+
+            self.message_bar = MessagesWidget(self)
+            self.message_bar.setSizePolicy(QSizePolicy.Preferred,
+                                           QSizePolicy.Preferred)
+            pb = QProgressBar(maximumWidth=120, minimum=0, maximum=100)
+            pb.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Ignored)
+            pb.setAttribute(Qt.WA_LayoutUsesWidgetRect)
+            pb.setAttribute(Qt.WA_MacMiniSize)
+            pb.hide()
+            sb.addPermanentWidget(pb)
+            sb.addPermanentWidget(self.message_bar)
+
+            def statechanged():
+                pb.setVisible(bool(self.processingState) or self.isBlocking())
+                if self.isBlocking() and not self.processingState:
+                    pb.setRange(0, 0)  # indeterminate pb
+                elif self.processingState:
+                    pb.setRange(0, 100)  # determinate pb
+
+            self.processingStateChanged.connect(statechanged)
+            self.blockingStateChanged.connect(statechanged)
+
+            @self.progressBarValueChanged.connect
+            def _(val):
+                pb.setValue(int(val))
+
+            # Reserve the bottom margins for the status bar
+            margins = self.layout().contentsMargins()
+            margins.setBottom(sb.sizeHint().height())
+            self.setContentsMargins(margins)
+
     def save_graph(self):
         """Save the graph with the name given in class attribute `graph_name`.
 
@@ -471,19 +521,6 @@ class OWWidget(QDialog, OWComponent, Report, ProgressBarMixin,
         self.raise_()
         self.activateWindow()
 
-    def send(self, signalName, value, id=None):
-        """
-        Send a `value` on the `signalName` widget output.
-
-        An output with `signalName` must be defined in the class ``outputs``
-        list.
-        """
-        if not any(s.name == signalName for s in self.outputs):
-            raise ValueError('{} is not a valid output signal for widget {}'.format(
-                signalName, self.name))
-        if self.signalManager is not None:
-            self.signalManager.send(self, signalName, value, id)
-
     def openContext(self, *a):
         """Open a new context corresponding to the given data.
 
@@ -820,12 +857,3 @@ Explicit = widget_description.Explicit
 #: to any input signal which can accept a subtype of the declared output
 #: type.
 Dynamic = widget_description.Dynamic
-
-InputSignal = widget_description.InputSignal
-OutputSignal = widget_description.OutputSignal
-
-
-class AttributeList(list):
-    """Signal type for lists of attributes (variables)
-    """
-    pass
